"""Exposes lower-level COM related methods."""
__all__ = [
    'create_IFileOperation',
    'parse_filename',
    'WIN32_FIND_DATAW',
]

from ctypes import POINTER, PyDLL, byref, c_bool, c_void_p, py_object
from ctypes.wintypes import DWORD, PWIN32_FIND_DATAW, WIN32_FIND_DATAW
from functools import wraps
from os import PathLike, fspath
from typing import Callable, ParamSpec, TypeAlias, TypeVar, cast

# pywin32
import pythoncom
import pywintypes

# comtypes
from comtypes import COMMETHOD, GUID, HRESULT, COMObject, IUnknown
from comtypes.hresult import S_OK
from win32com.shell import shell  # type: ignore (module *does* exist)

from .errors import FileOperatorError
from .flags import FileAttributeFlags
from .interfaces import IFileOperation

StrPath: TypeAlias = str | PathLike[str]
P_IUnknown = POINTER(IUnknown)  # type: ignore
P = ParamSpec('P')
T = TypeVar('T')


def create_IFileOperation() -> IFileOperation:
    """Create a IFileOperation instance."""
    ifo = pythoncom.CoCreateInstance(
        shell.CLSID_FileOperation, None, pythoncom.CLSCTX_ALL, shell.IID_IFileOperation
    )
    return cast(IFileOperation, ifo)


def parse_filename(path: StrPath, force: bool = False):
    """Parse a filename into an IShellItem2 instance. If `force` is True, then
    an IShellItem2 instance will be returned even if the file does not exist.
    """
    path = fspath(path)
    try:
        return shell.SHCreateItemFromParsingName(
            path, None, shell.IID_IShellItem2  # type: ignore
        )
    except pywintypes.com_error as e:
        if e.hresult == E_FILE_NOT_FOUND and force:  # type: ignore
            return shell.SHCreateItemFromParsingName(
                path, FOLDER_BIND_CTX, shell.IID_IShellItem2
            )
        else:
            raise


_PYCOM_PyObjectFromIUnknown = PyDLL(pythoncom.__file__).PyCom_PyObjectFromIUnknown
_PYCOM_PyObjectFromIUnknown.restype = py_object
_PYCOM_PyObjectFromIUnknown.argtypes = (P_IUnknown, c_void_p, c_bool)


def comtypes_to_pywin32(com_ptr, interface: type[IUnknown] = IUnknown):
    """Convert a comtypes pointer to a pythoncom PyI<interface> object."""
    # Cribbed from comtypes\test\test_win32com_interop.py
    # It uses pythoncom.dll to convert a COM pointer to a pythoncom pointer
    return _PYCOM_PyObjectFromIUnknown(com_ptr, byref(interface._iid_), True)


class IFileSysBindData(IUnknown):
    """Define the IFileSysBindData interface using comtypes, see
    https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nn-shobjidl_core-ifilesystembinddata
    """

    _iid_ = GUID('{01e18d10-4d8b-11d2-855d-006008059367}')
    _methods_ = [
        COMMETHOD([], HRESULT, 'SetFindData', (['in'], PWIN32_FIND_DATAW, 'pfd')),
        COMMETHOD(
            [], HRESULT, 'GetFindData', (['out'], POINTER(PWIN32_FIND_DATAW), 'ppfd')
        ),
    ]


class FileSysBindData(COMObject):
    """Implement the IFileSysBindData interface"""

    _com_interfaces_ = [IFileSysBindData]

    def SetFindData(self, pfd) -> int:
        self.find_data = pfd
        return S_OK

    def GetFindData(self, ppfd) -> int:
        ppfd.contents = self.find_data
        # ppfd = self.find_data
        return S_OK


E_FILE_NOT_FOUND = -2147024894


def create_bind_ctx(find_data: WIN32_FIND_DATAW):
    """Create an IBindCtx instance with a custom WIN32_FIND_DATA structure."""
    bind_data = FileSysBindData()
    bind_data.SetFindData(byref(find_data))  # type: ignore
    bind_ctx = pythoncom.CreateBindCtx()
    bind_ctx.RegisterObjectParam(
        'File System Bind Data', comtypes_to_pywin32(bind_data)
    )
    return bind_ctx


def convert_exceptions(callable: Callable[P, T]) -> Callable[P, T]:
    """Wraps a method to automatically convert pythoncom.com_error exceptions
    into FileOperatorError exceptions.
    """
    # TODO: Implement mapping to builtin Python exceptions for common
    # HRESULTs, e.g. E_ACCESSDENIED, E_OUTOFMEMORY, etc.
    @wraps(callable)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return callable(*args, **kwargs)
        except pythoncom.com_error as e:
            raise FileOperatorError(*e.args) from None

    return wrapped


FOLDER_BIND_CTX = create_bind_ctx(WIN32_FIND_DATAW(DWORD(FileAttributeFlags.DIRECTORY)))



__all__ = [
    'FileOperator',
]

from os import PathLike, fspath
from pathlib import Path
from typing import Iterable, TypeAlias

from .com import convert_exceptions, create_IFileOperation, parse_filename
from .errors import IFileOperationError
from .flags import FileOperationFlags

StrPath: TypeAlias = str | PathLike[str]


class FileOperator:
    """A context manager for performing file operations.  Not reentrant."""

    def __init__(
        self,
        parent=None,
        flags: FileOperationFlags | None = None,
        *,
        commit_on_exit: bool = False,
    ):
        """Create a FileOperator instance for managing file operations.  To be used with
        the `with` statement. Operations are scheduled first, then performed with the
        `commit` method, or optionally automatically when the with block is exited.

        The context manager is not reentrant.

        :param parent: A HANDLE to the window that should own any dialog boxes that are
            displayed. Also supports wx.Window instances.
        :param flags: The FileOperationFlags to use for the operations.
        :param commit_on_exit: If True (default: False), `commit` will be performed
            automatically when the with block is exited if no exceptions were raised.
        """
        self.ifo = create_IFileOperation()
        if parent is not None:
            try:
                parent = parent.GetHandle()  # wx.Window
            except AttributeError:
                pass
        if parent is not None:
            self.ifo.SetOwnerWindow(parent)
        if flags is not None:
            self.ifo.SetOperationFlags(int(flags))
        self.commit_on_exit = commit_on_exit
        self.entered = False

    def __enter__(self):
        if self.entered:
            raise IFileOperationError(f'{type(self).__name__} is not reentrant')
        else:
            self.entered = True
            return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.commit_on_exit and not exc_type:
            self.commit()
        # Delete the IFileOperation instance, so reentrance will raise an error
        del self.ifo

    @convert_exceptions
    def move_file(
        self, source: StrPath, destination: StrPath, new_name: str | None = None
    ) -> None:
        """Schedule to move a file from `source` to `destination`.  Optionally rename
        the file in the process. NOTE: Windows performs a move to a different logical
        drive as a copy and delete.

        :param source: The file to be moved.
        :param destination: The destination *directory* to move to.
        :param new_name: The new name for the file if desired.
        """
        self.ifo.MoveItem(
            parse_filename(source), parse_filename(destination, True), new_name
        )

    @convert_exceptions
    def move_files(self, sources: Iterable[StrPath], destination: StrPath) -> None:
        """Schedule to move the files `sources` to `destination`. NOTE: Windows performs
        a move to a different logical drive as a copy and delete.

        :param sources: And iterable of filenames to be moved.
        :param destination: The destination *directory* to move to.
        """
        srcs = [parse_filename(s) for s in sources]
        self.ifo.MoveItems(srcs, parse_filename(destination, True))

    @convert_exceptions
    def rename_file(
        self, source: StrPath, new_name: str, allow_move: bool = True
    ) -> None:
        """Schedule to rename `source` to `new_name`.  If `allow_move` is True and the
        destination is in a different directory, this will perform a move instead.

        :param source: The file to be renamed.
        :param new_name: The new name of the file.
        :param allow_move: If True, then converts a rename into a move when new_name is
            in a different directory than the source.
        """
        if allow_move:
            # Check if new_name refers to a file in a different directory
            dest_path = Path(new_name)
            if len((dest_parts := dest_path.parts)) == 1:
                # No directory on the new_name, all good for a rename
                pass
            elif Path(source).parts[:-1] != dest_parts[:-1]:
                # new_name wants the file to be in a different directory, do a move
                # instead
                self.move_file(source, dest_path.parent, dest_path.name)
                return
            else:
                # new_name includes the directory, but it's the same directory, so a
                # rename is fine
                new_name = dest_path.name
        self.ifo.RenameItem(parse_filename(source), new_name)

    @convert_exceptions
    def rename_files(self, sources: Iterable[StrPath], new_name: StrPath) -> None:
        """Schedule to rename the files `sources` to `new_name`.  All files will be
        renamed to the new name.

        :param sources: An iterable of filenames to be renamed.
        :param new_name: The new name of the files.
        """
        srcs = [parse_filename(s) for s in sources]
        self.ifo.RenameItems(srcs, fspath(new_name))

    @convert_exceptions
    def copy_file(
        self, source: StrPath, destination: StrPath, new_name: str | None = None
    ) -> None:
        """Schedule to copy a file from `source` to `destination`. Optionally rename the
        file in the process.

        :param source: The file to be copied.
        :param destination: The destination *directory* to copy to.
        :param new_name: Optional new name of the file.
        """
        self.ifo.CopyItem(
            parse_filename(source), parse_filename(destination, True), new_name
        )

    @convert_exceptions
    def copy_files(self, sources: Iterable[StrPath], destination: StrPath) -> None:
        """Schedule to copy the files `sources` to `destination`.

        :param sources: An iterable of filenames to be copied.
        :param destination: The destination *directory* to copy to.
        """
        srcs = [parse_filename(s) for s in sources]
        self.ifo.CopyItems(srcs, parse_filename(destination, True))

    @convert_exceptions
    def delete_file(self, source: StrPath) -> None:
        """Schedule to delete a file.

        :param source: The file to be deleted.
        """
        self.ifo.DeleteItem(parse_filename(source))

    @convert_exceptions
    def delete_files(self, sources: Iterable[StrPath]) -> None:
        """Schedule to delete the files `sources`.

        :param sources: An iterable of filenames to be deleted.
        """
        srcs = [parse_filename(s) for s in sources]
        self.ifo.DeleteItems(srcs)

    @convert_exceptions
    def commit(self):
        """Perform all scheduled file operations."""
        self.returncode = self.ifo.PerformOperations()
        self.aborted = self.ifo.GetAnyOperationsAborted()
